@@ -28,21 +28,6 @@
 #warning *****************************************************************
 #endif
 
-#if (CUDA_VERSION == 3000)
-#warning *****************************************************************
-#warning **** CUDA 3.0 detected. 
-#warning **** (Don't worry, that's not in itself bad.)
-#warning *****************************************************************
-#warning **** PyCUDA assumes that you are using the release (non-beta)
-#warning **** version. 3.0 beta will result in compile errors.
-#warning **** I apologize, but I have no way of checking whether I am
-#warning **** compiling against the beta--Nvidia cleverly changed the
-#warning **** API without bumping the version number.
-#warning **** You may try undefining CUDAPP_POST_30_BETA in 
-#warning **** src/cpp/cuda.hpp to get this to work.
-#warning *****************************************************************
-#endif
-
 
 
 
@@ -52,14 +37,20 @@
 
 
 
+// {{{ tracing and error guards
+
 #ifdef CUDAPP_TRACE_CUDA
   #define CUDAPP_PRINT_CALL_TRACE(NAME) \
     std::cerr << NAME << std::endl;
   #define CUDAPP_PRINT_CALL_TRACE_INFO(NAME, EXTRA_INFO) \
     std::cerr << NAME << " (" << EXTRA_INFO << ')' << std::endl;
+  #define CUDAPP_PRINT_ERROR_TRACE(NAME, CODE) \
+    if (CODE != CUDA_SUCCESS) \
+      std::cerr << NAME << " failed with code " << CODE << std::endl;
 #else
   #define CUDAPP_PRINT_CALL_TRACE(NAME) /*nothing*/
   #define CUDAPP_PRINT_CALL_TRACE_INFO(NAME, EXTRA_INFO) /*nothing*/
+  #define CUDAPP_PRINT_ERROR_TRACE(NAME, CODE) /*nothing*/
 #endif
 
 #define CUDAPP_CALL_GUARDED_THREADED_WITH_TRACE_INFO(NAME, ARGLIST, TRACE_INFO) \
@@ -78,6 +69,7 @@
     CUDAPP_PRINT_CALL_TRACE_INFO(#NAME, TRACE_INFO); \
     CUresult cu_status_code; \
     cu_status_code = NAME ARGLIST; \
+    CUDAPP_PRINT_ERROR_TRACE(#NAME, cu_status_code); \
     if (cu_status_code != CUDA_SUCCESS) \
       throw cuda::error(#NAME, cu_status_code);\
   }
@@ -89,6 +81,7 @@
     Py_BEGIN_ALLOW_THREADS \
       cu_status_code = NAME ARGLIST; \
     Py_END_ALLOW_THREADS \
+    CUDAPP_PRINT_ERROR_TRACE(#NAME, cu_status_code); \
     if (cu_status_code != CUDA_SUCCESS) \
       throw cuda::error(#NAME, cu_status_code);\
   }
@@ -98,6 +91,7 @@
     CUDAPP_PRINT_CALL_TRACE(#NAME); \
     CUresult cu_status_code; \
     cu_status_code = NAME ARGLIST; \
+    CUDAPP_PRINT_ERROR_TRACE(#NAME, cu_status_code); \
     if (cu_status_code != CUDA_SUCCESS) \
       throw cuda::error(#NAME, cu_status_code);\
   }
@@ -106,6 +100,7 @@
     CUDAPP_PRINT_CALL_TRACE(#NAME); \
     CUresult cu_status_code; \
     cu_status_code = NAME ARGLIST; \
+    CUDAPP_PRINT_ERROR_TRACE(#NAME, cu_status_code); \
     if (cu_status_code != CUDA_SUCCESS) \
       std::cerr \
         << "PyCUDA WARNING: a clean-up operation failed (dead context maybe?)" \
@@ -125,6 +120,8 @@
   // therefore its context has already been deleted. No need to harp
   // on the fact that we still thought there was cleanup to do.
 
+// }}}
+
 
 
 
@@ -132,9 +129,17 @@ namespace cuda
 {
   namespace py = boost::python;
 
+  typedef
+#if CUDA_VERSION >= 3020
+        size_t
+#else
+        unsigned int 
+#endif
+        pycuda_size_t;
 
 
 
+  // {{{ error reporting
   class error : public std::runtime_error
   {
     private:
@@ -201,10 +206,19 @@ namespace cuda
 #ifdef CUDAPP_POST_30_BETA
           case CUDA_ERROR_ECC_UNCORRECTABLE: return "ECC uncorrectable";
 #endif
+#endif
+#if CUDA_VERSION >= 3010
+          case CUDA_ERROR_UNSUPPORTED_LIMIT: return "unsupported limit";
 #endif
 
           case CUDA_ERROR_INVALID_SOURCE: return "invalid source";
           case CUDA_ERROR_FILE_NOT_FOUND: return "file not found";
+#if CUDA_VERSION >= 3010
+          case CUDA_ERROR_SHARED_OBJECT_SYMBOL_NOT_FOUND:
+            return "shared object symbol not found";
+          case CUDA_ERROR_SHARED_OBJECT_INIT_FAILED:
+            return "shared object init failed";
+#endif
 
           case CUDA_ERROR_INVALID_HANDLE: return "invalid handle";
 
@@ -217,9 +231,11 @@ namespace cuda
           case CUDA_ERROR_LAUNCH_TIMEOUT: return "launch timeout";
           case CUDA_ERROR_LAUNCH_INCOMPATIBLE_TEXTURING: return "launch incompatible texturing";
 
-#if CUDA_VERSION >= 3000
-          case CUDA_ERROR_POINTER_IS_64BIT: return "pointer is 64-bit";
-          case CUDA_ERROR_SIZE_IS_64BIT: return "size is 64-bit";
+#if (CUDA_VERSION >= 3000) && (CUDA_VERSION < 3020)
+          case CUDA_ERROR_POINTER_IS_64BIT:
+             return "attempted to retrieve 64-bit pointer via 32-bit api function";
+          case CUDA_ERROR_SIZE_IS_64BIT:
+             return "attempted to retrieve 64-bit size via 32-bit api function";
 #endif
 
           case CUDA_ERROR_UNKNOWN: return "unknown";
@@ -230,22 +246,22 @@ namespace cuda
   };
 
   struct cannot_activate_out_of_thread_context : public std::logic_error
-  { 
+  {
     cannot_activate_out_of_thread_context(std::string const &w)
       : std::logic_error(w)
     { }
   };
 
   struct cannot_activate_dead_context : public std::logic_error
-  { 
+  {
     cannot_activate_dead_context(std::string const &w)
       : std::logic_error(w)
     { }
   };
 
+  // }}}
 
-
-  // version query ------------------------------------------------------------
+  // {{{ version query ------------------------------------------------------------
 #if CUDA_VERSION >= 2020
   inline int get_driver_version()
   {
@@ -254,11 +270,9 @@ namespace cuda
     return result;
   }
 #endif
+  // }}}
 
-
-
-
-  // device -------------------------------------------------------------------
+  // {{{ device
   class context;
 
   class device
@@ -292,9 +306,10 @@ namespace cuda
         return py::make_tuple(major, minor);
       }
 
-      unsigned int total_memory()
+      pycuda_size_t total_memory()
       {
-        unsigned int bytes;
+        pycuda_size_t bytes;
+
         CUDAPP_CALL_GUARDED(cuDeviceTotalMem, (&bytes, m_device));
         return bytes;
       }
@@ -329,23 +344,22 @@ namespace cuda
   };
 
   inline
-  void init(unsigned int flags) 
-  { 
-    CUDAPP_CALL_GUARDED(cuInit, (flags)); 
+  void init(unsigned int flags)
+  {
+    CUDAPP_CALL_GUARDED(cuInit, (flags));
   }
 
   inline
   device *make_device(int ordinal)
-  { 
+  {
     CUdevice result;
-    CUDAPP_CALL_GUARDED(cuDeviceGet, (&result, ordinal)); 
+    CUDAPP_CALL_GUARDED(cuDeviceGet, (&result, ordinal));
     return new device(result);
   }
 
+  // }}}
 
-
-
-  // context ------------------------------------------------------------------
+  // {{{ context
   /* A word on context management: We don't let CUDA's context stack get more
    * than one deep. CUDA only supports pushing floating contexts. We may wish
    * to push contexts that are already active at a deeper stack level, so we
@@ -353,8 +367,8 @@ namespace cuda
    */
 
   // for friend decl
-  namespace gl { 
-    boost::shared_ptr<context> 
+  namespace gl {
+    boost::shared_ptr<context>
         make_gl_context(device const &dev, unsigned int flags);
   }
 
@@ -407,7 +421,7 @@ namespace cuda
 
     public:
       context(CUcontext ctx)
-        : m_context(ctx), m_valid(true), m_use_count(1), 
+        : m_context(ctx), m_valid(true), m_use_count(1),
         m_thread(boost::this_thread::get_id())
       { }
 
@@ -468,7 +482,7 @@ namespace cuda
             }
             else
             {
-              // In all likelihood, this context's managing thread has exited, and 
+              // In all likelihood, this context's managing thread has exited, and
               // therefore this context has already been deleted. No need to harp
               // on the fact that we still thought there was cleanup to do.
 
@@ -483,7 +497,7 @@ namespace cuda
             boost::shared_ptr<context> new_active = current_context(this);
             if (new_active.get())
             {
-              CUDAPP_CALL_GUARDED(cuCtxPushCurrent, (new_active->m_context)); 
+              CUDAPP_CALL_GUARDED(cuCtxPushCurrent, (new_active->m_context));
             }
           }
         }
@@ -493,9 +507,9 @@ namespace cuda
       }
 
       static device get_device()
-      { 
+      {
         CUdevice dev;
-        CUDAPP_CALL_GUARDED(cuCtxGetDevice, (&dev)); 
+        CUDAPP_CALL_GUARDED(cuCtxGetDevice, (&dev));
         return device(dev);
       }
 
@@ -506,12 +520,12 @@ namespace cuda
         if (!context_stack::get().empty())
         {
           CUcontext popped;
-          CUDAPP_CALL_GUARDED(cuCtxPopCurrent, (&popped)); 
+          CUDAPP_CALL_GUARDED(cuCtxPopCurrent, (&popped));
         }
       }
 
       static void pop()
-      { 
+      {
         prepare_context_switch();
         context_stack &ctx_stack = context_stack::get();
 
@@ -529,7 +543,7 @@ namespace cuda
 
         current = current_context();
         if (current)
-          CUDAPP_CALL_GUARDED(cuCtxPushCurrent, (current_context()->m_context)); 
+          CUDAPP_CALL_GUARDED(cuCtxPushCurrent, (current_context()->m_context));
       }
 #else
       static void prepare_context_switch() { }
@@ -546,7 +560,7 @@ namespace cuda
             return boost::shared_ptr<context>();
 
           boost::shared_ptr<context> result(context_stack::get().top());
-          if (result.get() != except 
+          if (result.get() != except
               && result->is_valid())
           {
             // good, weak pointer didn't expire
@@ -558,9 +572,44 @@ namespace cuda
         }
       }
 
+#if CUDA_VERSION >= 3010
+      static void set_limit(CUlimit limit, size_t value)
+      {
+        CUDAPP_CALL_GUARDED(cuCtxSetLimit, (limit, value));
+      }
+
+      static size_t get_limit(CUlimit limit)
+      {
+        size_t value;
+        CUDAPP_CALL_GUARDED(cuCtxGetLimit, (&value, limit));
+        return value;
+      }
+#endif
+
+#if CUDA_VERSION >= 3020
+      static CUfunc_cache get_cache_config()
+      {
+        CUfunc_cache value;
+        CUDAPP_CALL_GUARDED(cuCtxGetCacheConfig, (&value));
+        return value;
+      }
+
+      static void set_cache_config(CUfunc_cache cc)
+      {
+        CUDAPP_CALL_GUARDED(cuCtxSetCacheConfig, (cc));
+      }
+
+      unsigned int get_api_version()
+      {
+        unsigned int value;
+        CUDAPP_CALL_GUARDED(cuCtxGetApiVersion, (m_context, &value));
+        return value;
+      }
+#endif
+
       friend class device;
       friend void context_push(boost::shared_ptr<context> ctx);
-      friend boost::shared_ptr<context> 
+      friend boost::shared_ptr<context>
           gl::make_gl_context(device const &dev, unsigned int flags);
   };
 
@@ -585,10 +634,10 @@ namespace cuda
 #if CUDA_VERSION >= 2000
   inline
   void context_push(boost::shared_ptr<context> ctx)
-  { 
+  {
     context::prepare_context_switch();
 
-    CUDAPP_CALL_GUARDED(cuCtxPushCurrent, (ctx->m_context)); 
+    CUDAPP_CALL_GUARDED(cuCtxPushCurrent, (ctx->m_context));
     context_stack::get().push(ctx);
     ++ctx->m_use_count;
   }
@@ -601,7 +650,7 @@ namespace cuda
   {
     if (!m_stack.empty())
     {
-      std::cerr 
+      std::cerr
         << "-------------------------------------------------------------------" << std::endl
         << "PyCUDA ERROR: The context stack was not empty upon module cleanup." << std::endl
         << "-------------------------------------------------------------------" << std::endl
@@ -664,7 +713,7 @@ namespace cuda
     public:
       scoped_context_activation(boost::shared_ptr<context> ctx)
         : m_context(ctx)
-      { 
+      {
         if (!m_context->is_valid())
           throw cuda::cannot_activate_dead_context(
               "cannot activate dead context");
@@ -694,7 +743,9 @@ namespace cuda
 
   };
 
-  // streams ------------------------------------------------------------------
+  // }}}
+
+  // {{{ stream
   class stream : public boost::noncopyable, public context_dependent
   {
     private:
@@ -705,11 +756,11 @@ namespace cuda
       { CUDAPP_CALL_GUARDED(cuStreamCreate, (&m_stream, flags)); }
 
       ~stream()
-      { 
+      {
         try
         {
           scoped_context_activation ca(get_context());
-          CUDAPP_CALL_GUARDED_CLEANUP(cuStreamDestroy, (m_stream)); 
+          CUDAPP_CALL_GUARDED_CLEANUP(cuStreamDestroy, (m_stream));
         }
         CUDAPP_CATCH_CLEANUP_ON_DEAD_CONTEXT(stream);
       }
@@ -721,27 +772,25 @@ namespace cuda
       { return m_stream; }
 
       bool is_done() const
-      { 
-#ifdef TRACE_CUDA
-        std::cerr << "cuStreamQuery" << std::endl;
-#endif
+      {
+        CUDAPP_PRINT_CALL_TRACE("cuStreamQuery");
         CUresult result = cuStreamQuery(m_stream);
         switch (result)
         {
-          case CUDA_SUCCESS: 
+          case CUDA_SUCCESS:
             return true;
-          case CUDA_ERROR_NOT_READY: 
+          case CUDA_ERROR_NOT_READY:
             return false;
           default:
+            CUDAPP_PRINT_ERROR_TRACE("cuStreamQuery", result);
             throw error("cuStreamQuery", result);
         }
       }
   };
 
+  // }}}
 
-
-
-  // arrays -------------------------------------------------------------------
+  // {{{ array
   class array : public boost::noncopyable, public context_dependent
   {
     private:
@@ -773,7 +822,7 @@ namespace cuda
           try
           {
             scoped_context_activation ca(get_context());
-            CUDAPP_CALL_GUARDED_CLEANUP(cuArrayDestroy, (m_array)); 
+            CUDAPP_CALL_GUARDED_CLEANUP(cuArrayDestroy, (m_array));
           }
           CUDAPP_CATCH_CLEANUP_ON_DEAD_CONTEXT(array);
 
@@ -802,10 +851,9 @@ namespace cuda
       { return m_array; }
   };
 
+  // }}}
 
-
-
-  // texture reference --------------------------------------------------------
+  // {{{ texture reference
   class module;
 
   class texture_reference : public  boost::noncopyable
@@ -828,10 +876,10 @@ namespace cuda
       { }
 
       ~texture_reference()
-      { 
+      {
         if (m_managed)
         {
-          CUDAPP_CALL_GUARDED_CLEANUP(cuTexRefDestroy, (m_texref)); 
+          CUDAPP_CALL_GUARDED_CLEANUP(cuTexRefDestroy, (m_texref));
         }
       }
 
@@ -842,17 +890,17 @@ namespace cuda
       { return m_texref; }
 
       void set_array(boost::shared_ptr<array> ary)
-      { 
-        CUDAPP_CALL_GUARDED(cuTexRefSetArray, (m_texref, 
-            ary->handle(), CU_TRSA_OVERRIDE_FORMAT)); 
+      {
+        CUDAPP_CALL_GUARDED(cuTexRefSetArray, (m_texref,
+            ary->handle(), CU_TRSA_OVERRIDE_FORMAT));
         m_array = ary;
       }
 
-      unsigned int set_address(CUdeviceptr dptr, unsigned int bytes, bool allow_offset=false)
-      { 
-        unsigned int byte_offset;
+      pycuda_size_t set_address(CUdeviceptr dptr, unsigned int bytes, bool allow_offset=false)
+      {
+        pycuda_size_t byte_offset;
         CUDAPP_CALL_GUARDED(cuTexRefSetAddress, (&byte_offset,
-              m_texref, dptr, bytes)); 
+              m_texref, dptr, bytes));
 
         if (!allow_offset && byte_offset != 0)
           throw cuda::error("texture_reference::set_address", CUDA_ERROR_INVALID_VALUE,
@@ -863,10 +911,10 @@ namespace cuda
       }
 
 #if CUDA_VERSION >= 2020
-      void set_address_2d(CUdeviceptr dptr, 
+      void set_address_2d(CUdeviceptr dptr,
           const CUDA_ARRAY_DESCRIPTOR &descr, unsigned int pitch)
-      { 
-        CUDAPP_CALL_GUARDED(cuTexRefSetAddress2D, (m_texref, &descr, dptr, pitch)); 
+      {
+        CUDAPP_CALL_GUARDED(cuTexRefSetAddress2D, (m_texref, &descr, dptr, pitch));
       }
 #endif
 
@@ -924,10 +972,50 @@ namespace cuda
       }
   };
 
+  // }}}
 
+  // {{{ surface reference
+#if CUDA_VERSION >= 3010
+  class module;
 
+  class surface_reference : public  boost::noncopyable
+  {
+    private:
+      CUsurfref m_surfref;
 
-  // module -------------------------------------------------------------------
+      // life support for array and module
+      boost::shared_ptr<array> m_array;
+      boost::shared_ptr<module> m_module;
+
+    public:
+      surface_reference(CUsurfref sr)
+        : m_surfref(sr)
+      { }
+
+      void set_module(boost::shared_ptr<module> mod)
+      { m_module = mod; }
+
+      CUsurfref handle() const
+      { return m_surfref; }
+
+      void set_array(boost::shared_ptr<array> ary, unsigned int flags)
+      {
+        CUDAPP_CALL_GUARDED(cuSurfRefSetArray, (m_surfref, ary->handle(), flags));
+        m_array = ary;
+      }
+
+      array *get_array()
+      {
+        CUarray result;
+        CUDAPP_CALL_GUARDED(cuSurfRefGetArray, (&result, m_surfref));
+        return new array(result, false);
+      }
+  };
+#endif
+
+  // }}}
+
+  // {{{ module
   class function;
 
   class module : public boost::noncopyable, public context_dependent
@@ -957,7 +1045,7 @@ namespace cuda
       py::tuple get_global(const char *name)
       {
         CUdeviceptr devptr;
-        unsigned int bytes;
+        pycuda_size_t bytes;
         CUDAPP_CALL_GUARDED(cuModuleGetGlobal, (&devptr, &bytes, m_module, name));
         return py::make_tuple(devptr, bytes);
       }
@@ -972,7 +1060,8 @@ namespace cuda
   }
 
   inline
-  texture_reference *module_get_texref(boost::shared_ptr<module> mod, const char *name)
+  texture_reference *module_get_texref(
+      boost::shared_ptr<module> mod, const char *name)
   {
     CUtexref tr;
     CUDAPP_CALL_GUARDED(cuModuleGetTexRef, (&tr, mod->handle(), name));
@@ -982,10 +1071,23 @@ namespace cuda
     return result.release();
   }
 
+#if CUDA_VERSION >= 3010
+  inline
+  surface_reference *module_get_surfref(
+      boost::shared_ptr<module> mod, const char *name)
+  {
+    CUsurfref sr;
+    CUDAPP_CALL_GUARDED(cuModuleGetSurfRef, (&sr, mod->handle(), name));
+    std::auto_ptr<surface_reference> result(
+        new surface_reference(sr));
+    result->set_module(mod);
+    return result.release();
+  }
+#endif
 
+  // }}}
 
-
-  // function -----------------------------------------------------------------
+  // {{{ function
   class function
   {
     private:
@@ -998,56 +1100,56 @@ namespace cuda
       { }
 
       void set_block_shape(int x, int y, int z)
-      { 
+      {
         CUDAPP_CALL_GUARDED_WITH_TRACE_INFO(
-            cuFuncSetBlockShape, (m_function, x, y, z), m_symbol); 
+            cuFuncSetBlockShape, (m_function, x, y, z), m_symbol);
       }
       void set_shared_size(unsigned int bytes)
-      { 
+      {
         CUDAPP_CALL_GUARDED_WITH_TRACE_INFO(
-            cuFuncSetSharedSize, (m_function, bytes), m_symbol); 
+            cuFuncSetSharedSize, (m_function, bytes), m_symbol);
       }
 
       void param_set_size(unsigned int bytes)
-      { 
+      {
         CUDAPP_CALL_GUARDED_WITH_TRACE_INFO(
-            cuParamSetSize, (m_function, bytes), m_symbol); 
+            cuParamSetSize, (m_function, bytes), m_symbol);
       }
       void param_set(int offset, unsigned int value)
-      { 
+      {
         CUDAPP_CALL_GUARDED_WITH_TRACE_INFO(
-            cuParamSeti, (m_function, offset, value), m_symbol); 
+            cuParamSeti, (m_function, offset, value), m_symbol);
       }
       void param_set(int offset, float value)
-      { 
+      {
         CUDAPP_CALL_GUARDED_WITH_TRACE_INFO(
-          cuParamSetf, (m_function, offset, value), m_symbol); 
+          cuParamSetf, (m_function, offset, value), m_symbol);
       }
       void param_setv(int offset, void *buf, unsigned long len)
-      { 
+      {
         CUDAPP_CALL_GUARDED_WITH_TRACE_INFO(
-            cuParamSetv, (m_function, offset, buf, len), m_symbol); 
+            cuParamSetv, (m_function, offset, buf, len), m_symbol);
       }
       void param_set_texref(const texture_reference &tr)
-      { 
-        CUDAPP_CALL_GUARDED_WITH_TRACE_INFO(cuParamSetTexRef, (m_function, 
-            CU_PARAM_TR_DEFAULT, tr.handle()), m_symbol); 
+      {
+        CUDAPP_CALL_GUARDED_WITH_TRACE_INFO(cuParamSetTexRef, (m_function,
+            CU_PARAM_TR_DEFAULT, tr.handle()), m_symbol);
       }
 
       void launch()
-      { 
+      {
         CUDAPP_CALL_GUARDED_THREADED_WITH_TRACE_INFO(
-            cuLaunch, (m_function), m_symbol); 
+            cuLaunch, (m_function), m_symbol);
       }
       void launch_grid(int grid_width, int grid_height)
-      { 
+      {
         CUDAPP_CALL_GUARDED_THREADED_WITH_TRACE_INFO(
-          cuLaunchGrid, (m_function, grid_width, grid_height), m_symbol); 
+          cuLaunchGrid, (m_function, grid_width, grid_height), m_symbol);
       }
       void launch_grid_async(int grid_width, int grid_height, const stream &s)
-      { 
+      {
         CUDAPP_CALL_GUARDED_THREADED_WITH_TRACE_INFO(
-            cuLaunchGridAsync, (m_function, grid_width, grid_height, s.handle()), 
+            cuLaunchGridAsync, (m_function, grid_width, grid_height, s.handle()),
             m_symbol);
       }
 
@@ -1078,19 +1180,18 @@ namespace cuda
     return function(func, name);
   }
 
+  // }}}
 
-
-
-  // device memory ------------------------------------------------------------
+  // {{{ device memory
   inline
   py::tuple mem_get_info()
   {
-    unsigned int free, total;
+    pycuda_size_t free, total;
     CUDAPP_CALL_GUARDED(cuMemGetInfo, (&free, &total));
     return py::make_tuple(free, total);
   }
 
-  inline 
+  inline
   CUdeviceptr mem_alloc(unsigned long bytes)
   {
     CUdeviceptr devptr;
@@ -1098,12 +1199,24 @@ namespace cuda
     return devptr;
   }
 
-  inline 
+  inline
   void mem_free(CUdeviceptr devptr)
   {
     CUDAPP_CALL_GUARDED_CLEANUP(cuMemFree, (devptr));
   }
 
+  // A class the user can override to make device_allocation-
+  // workalikes.
+
+  class pointer_holder_base
+  {
+    public:
+      virtual ~pointer_holder_base() { }
+      virtual CUdeviceptr get_pointer() = 0;
+      operator CUdeviceptr()
+      { return get_pointer(); }
+  };
+
   class device_allocation : public boost::noncopyable, public context_dependent
   {
     private:
@@ -1140,7 +1253,7 @@ namespace cuda
         if (m_valid)
           free();
       }
-      
+
       operator CUdeviceptr() const
       { return m_devptr; }
   };
@@ -1150,7 +1263,7 @@ namespace cuda
         unsigned int width, unsigned int height, unsigned int access_size)
   {
     CUdeviceptr devptr;
-    unsigned int pitch;
+    pycuda_size_t pitch;
     CUDAPP_CALL_GUARDED(cuMemAllocPitch, (&devptr, &pitch, width, height, access_size));
     da = std::auto_ptr<device_allocation>(new device_allocation(devptr));
     return pitch;
@@ -1160,7 +1273,7 @@ namespace cuda
   py::tuple mem_get_address_range(CUdeviceptr ptr)
   {
     CUdeviceptr base;
-    unsigned int size;
+    pycuda_size_t size;
     CUDAPP_CALL_GUARDED(cuMemGetAddressRange, (&base, &size, ptr));
     return py::make_tuple(base, size);
   }
@@ -1177,15 +1290,14 @@ namespace cuda
 
   inline
   void memcpy_atoa(
-      array const &dst, unsigned int dst_index, 
-      array const &src, unsigned int src_index, 
+      array const &dst, unsigned int dst_index,
+      array const &src, unsigned int src_index,
       unsigned int len)
   { CUDAPP_CALL_GUARDED_THREADED(cuMemcpyAtoA, (dst.handle(), dst_index, src.handle(), src_index, len)); }
 
+  // }}}
 
-
-
-  // structured memcpy --------------------------------------------------------
+  // {{{ structured memcpy
 #if PY_VERSION_HEX >= 0x02050000
   typedef Py_ssize_t PYCUDA_BUFFER_SIZE_T;
 #else
@@ -1293,8 +1405,9 @@ namespace cuda
   };
 #endif
 
+  // }}}
 
-  // host memory --------------------------------------------------------------
+  // {{{ host memory
   inline void *mem_alloc_host(unsigned int size, unsigned flags=0)
   {
     void *m_data;
@@ -1329,9 +1442,9 @@ namespace cuda
       { }
 
       ~host_allocation()
-      { 
+      {
         if (m_valid)
-          free(); 
+          free();
       }
 
       void free()
@@ -1341,7 +1454,7 @@ namespace cuda
           try
           {
             scoped_context_activation ca(get_context());
-            mem_free_host(m_data); 
+            mem_free_host(m_data);
           }
           CUDAPP_CATCH_CLEANUP_ON_DEAD_CONTEXT(host_allocation);
 
@@ -1351,7 +1464,7 @@ namespace cuda
         else
           throw cuda::error("host_allocation::free", CUDA_ERROR_INVALID_HANDLE);
       }
-      
+
       void *data()
       { return m_data; }
 
@@ -1364,12 +1477,20 @@ namespace cuda
       }
 #endif
 
+#if CUDA_VERSION >= 3020
+      unsigned int get_flags()
+      {
+        unsigned int flags;
+        CUDAPP_CALL_GUARDED(cuMemHostGetFlags, (&flags, m_data));
+        return flags;
+      }
+#endif
+
   };
 
+  // }}}
 
-
-
-  // events -------------------------------------------------------------------
+  // {{{ event
   class event : public boost::noncopyable, public context_dependent
   {
     private:
@@ -1380,17 +1501,17 @@ namespace cuda
       { CUDAPP_CALL_GUARDED(cuEventCreate, (&m_event, flags)); }
 
       ~event()
-      { 
+      {
         try
         {
           scoped_context_activation ca(get_context());
-          CUDAPP_CALL_GUARDED_CLEANUP(cuEventDestroy, (m_event)); 
+          CUDAPP_CALL_GUARDED_CLEANUP(cuEventDestroy, (m_event));
         }
         CUDAPP_CATCH_CLEANUP_ON_DEAD_CONTEXT(event);
       }
 
       event *record(py::object stream_py)
-      { 
+      {
         CUstream s_handle;
         if (stream_py.ptr() != Py_None)
         {
@@ -1400,29 +1521,29 @@ namespace cuda
         else
           s_handle = 0;
 
-        CUDAPP_CALL_GUARDED(cuEventRecord, (m_event, s_handle)); 
+        CUDAPP_CALL_GUARDED(cuEventRecord, (m_event, s_handle));
         return this;
       }
 
       event *synchronize()
-      { 
-        CUDAPP_CALL_GUARDED_THREADED(cuEventSynchronize, (m_event)); 
+      {
+        CUDAPP_CALL_GUARDED_THREADED(cuEventSynchronize, (m_event));
         return this;
       }
 
       bool query() const
-      { 
-#ifdef TRACE_CUDA
-        std::cerr << "cuEventQuery" << std::endl;
-#endif
+      {
+        CUDAPP_PRINT_CALL_TRACE("cuEventQuery");
+
         CUresult result = cuEventQuery(m_event);
         switch (result)
         {
-          case CUDA_SUCCESS: 
+          case CUDA_SUCCESS:
             return true;
-          case CUDA_ERROR_NOT_READY: 
+          case CUDA_ERROR_NOT_READY:
             return false;
           default:
+            CUDAPP_PRINT_ERROR_TRACE("cuEventQuery", result);
             throw error("cuEventQuery", result);
         }
       }
@@ -1441,9 +1562,12 @@ namespace cuda
         return result;
       }
   };
+
+  // }}}
 }
 
 
 
 
 #endif
+// vim: foldmethod=marker
