@@ -26,6 +26,8 @@ using boost::shared_ptr;
 
 namespace
 {
+  // {{{ error handling
+
   py::handle<> 
     CudaError, 
     CudaMemoryError, 
@@ -62,8 +64,7 @@ namespace
       PyErr_SetString(CudaLogicError.get(), err.what());
   }
 
-
-
+  // }}}
 
   py::tuple cuda_version()
   {
@@ -80,6 +81,8 @@ namespace
 
 
 
+  // {{{ "python-aware" wrappers
+
   py::object device_get_attribute(device const &dev, CUdevice_attribute attr)
   {
 #if CUDA_VERSION >= 2020
@@ -97,6 +100,17 @@ namespace
     return new device_allocation(pycuda::mem_alloc_gc(bytes));
   }
 
+  class pointer_holder_base_wrap 
+    : public pointer_holder_base, 
+    public py::wrapper<pointer_holder_base>
+  {
+    public:
+      CUdeviceptr get_pointer()
+      {
+        return this->get_override("get_pointer")();
+      }
+  };
+
   py::tuple mem_alloc_pitch_wrap(
       unsigned int width, unsigned int height, unsigned int access_size)
   {
@@ -107,8 +121,7 @@ namespace
         handle_from_new_ptr(da.release()), pitch);
   }
 
-
-
+  // {{{ memory set
 
   void  py_memset_d8(CUdeviceptr dst, unsigned char uc, unsigned int n ) 
   { CUDAPP_CALL_GUARDED_THREADED(cuMemsetD8, (dst, uc, n )); }
@@ -129,167 +142,9 @@ namespace
       unsigned int ui, unsigned int width, unsigned int height )
   { CUDAPP_CALL_GUARDED_THREADED(cuMemsetD2D32, (dst, dst_pitch, ui, width, height)); }
 
-  void  py_memcpy_dtod(CUdeviceptr dst, CUdeviceptr src, 
-      unsigned int byte_count)
-  { CUDAPP_CALL_GUARDED_THREADED(cuMemcpyDtoD, (dst, src, byte_count)); }
-
-
-
-
-#if CUDA_VERSION >= 3000
-  void  py_memcpy_dtod_async(CUdeviceptr dst, CUdeviceptr src, 
-      unsigned int byte_count, py::object stream_py)
-  {
-    CUstream s_handle;
-    if (stream_py.ptr() != Py_None)
-    {
-      const stream &s = py::extract<const stream &>(stream_py);
-      s_handle = s.handle();
-    }
-    else
-      s_handle = 0;
-
-    CUDAPP_CALL_GUARDED_THREADED(cuMemcpyDtoDAsync, 
-        (dst, src, byte_count, s_handle)); 
-  }
-#endif
-
-
-
-
-  void function_param_setv(function &f, int offset, py::object buffer)
-  { 
-    const void *buf;
-    PYCUDA_BUFFER_SIZE_T len;
-    if (PyObject_AsReadBuffer(buffer.ptr(), &buf, &len))
-      throw py::error_already_set();
-    f.param_setv(offset, const_cast<void *>(buf), len);
-  }
-
-
-
-
-  module *module_from_buffer(py::object buffer, py::object py_options, 
-      py::object message_handler)
-  {
-    const char *mod_buf;
-    PYCUDA_BUFFER_SIZE_T len;
-    if (PyObject_AsCharBuffer(buffer.ptr(), &mod_buf, &len))
-      throw py::error_already_set();
-    CUmodule mod;
-
-#if CUDA_VERSION >= 2010
-    const unsigned buf_size = 32768;
-    char info_buf[buf_size], error_buf[buf_size];
-
-    std::vector<CUjit_option> options;
-    std::vector<void *> option_values;
-
-#define ADD_OPTION_PTR(KEY, PTR) \
-    { \
-      options.push_back(KEY); \
-      option_values.push_back(PTR); \
-    }
-
-    ADD_OPTION_PTR(CU_JIT_INFO_LOG_BUFFER, info_buf);
-    ADD_OPTION_PTR(CU_JIT_INFO_LOG_BUFFER_SIZE_BYTES, (void *) buf_size);
-    ADD_OPTION_PTR(CU_JIT_ERROR_LOG_BUFFER, error_buf);
-    ADD_OPTION_PTR(CU_JIT_ERROR_LOG_BUFFER_SIZE_BYTES, (void *) buf_size);
-
-    PYTHON_FOREACH(key_value, py_options)
-      ADD_OPTION_PTR(
-          py::extract<CUjit_option>(key_value[0]), 
-          (void *) py::extract<int>(key_value[1])());
-#undef ADD_OPTION
-    
-    CUDAPP_PRINT_CALL_TRACE("cuModuleLoadDataEx");
-    CUresult cu_status_code; \
-    cu_status_code = cuModuleLoadDataEx(&mod, mod_buf, options.size(), 
-         const_cast<CUjit_option *>(&*options.begin()),
-         const_cast<void **>(&*option_values.begin()));
-
-    size_t info_buf_size = size_t(option_values[1]);
-    size_t error_buf_size = size_t(option_values[3]);
-
-    if (message_handler != py::object())
-      message_handler(cu_status_code == CUDA_SUCCESS,
-          std::string(info_buf, info_buf_size),
-          std::string(error_buf, error_buf_size));
-
-    if (cu_status_code != CUDA_SUCCESS)
-      throw cuda::error("cuModuleLoadDataEx", cu_status_code, 
-          std::string(error_buf, error_buf_size).c_str());
-#else
-    if (py::len(py_options))
-      throw cuda::error("module_from_buffer", CUDA_ERROR_INVALID_VALUE,
-          "non-empty options argument only supported on CUDA 2.1 and newer");
-
-    CUDAPP_CALL_GUARDED(cuModuleLoadData, (&mod, mod_buf));
-#endif
-
-    return new module(mod);
-  }
-
-
-
-
-  PyObject *device_allocation_to_long(device_allocation const &da)
-  {
-    return PyLong_FromUnsignedLong((CUdeviceptr) da);
-  }
-
-
-
-
-  py::handle<> pagelocked_empty(py::object shape, py::object dtype, 
-      py::object order_py, unsigned mem_flags)
-  {
-    PyArray_Descr *tp_descr;
-    if (PyArray_DescrConverter(dtype.ptr(), &tp_descr) != NPY_SUCCEED)
-      throw py::error_already_set();
-
-    py::extract<npy_intp> shape_as_int(shape);
-    std::vector<npy_intp> dims;
-
-    if (shape_as_int.check())
-      dims.push_back(shape_as_int());
-    else
-      std::copy(
-          py::stl_input_iterator<npy_intp>(shape),
-          py::stl_input_iterator<npy_intp>(),
-          back_inserter(dims));
-
-    std::auto_ptr<host_allocation> alloc(
-        new host_allocation(
-          tp_descr->elsize*pycuda::size_from_dims(dims.size(), &dims.front()),
-          mem_flags)
-        );
-
-    NPY_ORDER order = PyArray_CORDER;
-    PyArray_OrderConverter(order_py.ptr(), &order);
-
-    int ary_flags = 0;
-    if (order == PyArray_FORTRANORDER)
-      ary_flags |= NPY_FARRAY;
-    else if (order == PyArray_CORDER)
-      ary_flags |= NPY_CARRAY;
-    else
-      throw std::runtime_error("unrecognized order specifier");
-
-    py::handle<> result = py::handle<>(PyArray_NewFromDescr(
-        &PyArray_Type, tp_descr,
-        dims.size(), &dims.front(), /*strides*/ NULL,
-        alloc->data(), ary_flags, /*obj*/NULL));
-
-    py::handle<> alloc_py(handle_from_new_ptr(alloc.release()));
-    PyArray_BASE(result.get()) = alloc_py.get();
-    Py_INCREF(alloc_py.get());
-
-    return result;
-  }
-
-
+  // }}}
 
+  // {{{ memory copies
 
   void py_memcpy_htod(CUdeviceptr dst, py::object src)
   {
@@ -384,6 +239,174 @@ namespace
     CUDAPP_CALL_GUARDED(cuMemcpyAtoH, (buf, ary.handle(), index, len));
   }
 
+
+
+
+  void  py_memcpy_dtod(CUdeviceptr dst, CUdeviceptr src, 
+      unsigned int byte_count)
+  { CUDAPP_CALL_GUARDED_THREADED(cuMemcpyDtoD, (dst, src, byte_count)); }
+
+
+
+
+#if CUDA_VERSION >= 3000
+  void  py_memcpy_dtod_async(CUdeviceptr dst, CUdeviceptr src, 
+      unsigned int byte_count, py::object stream_py)
+  {
+    CUstream s_handle;
+    if (stream_py.ptr() != Py_None)
+    {
+      const stream &s = py::extract<const stream &>(stream_py);
+      s_handle = s.handle();
+    }
+    else
+      s_handle = 0;
+
+    CUDAPP_CALL_GUARDED_THREADED(cuMemcpyDtoDAsync, 
+        (dst, src, byte_count, s_handle)); 
+  }
+#endif
+
+  // }}}
+
+  void function_param_setv(function &f, int offset, py::object buffer)
+  { 
+    const void *buf;
+    PYCUDA_BUFFER_SIZE_T len;
+    if (PyObject_AsReadBuffer(buffer.ptr(), &buf, &len))
+      throw py::error_already_set();
+    f.param_setv(offset, const_cast<void *>(buf), len);
+  }
+
+
+
+
+  // {{{ module_from_buffer
+
+  module *module_from_buffer(py::object buffer, py::object py_options, 
+      py::object message_handler)
+  {
+    const char *mod_buf;
+    PYCUDA_BUFFER_SIZE_T len;
+    if (PyObject_AsCharBuffer(buffer.ptr(), &mod_buf, &len))
+      throw py::error_already_set();
+    CUmodule mod;
+
+#if CUDA_VERSION >= 2010
+    const unsigned buf_size = 32768;
+    char info_buf[buf_size], error_buf[buf_size];
+
+    std::vector<CUjit_option> options;
+    std::vector<void *> option_values;
+
+#define ADD_OPTION_PTR(KEY, PTR) \
+    { \
+      options.push_back(KEY); \
+      option_values.push_back(PTR); \
+    }
+
+    ADD_OPTION_PTR(CU_JIT_INFO_LOG_BUFFER, info_buf);
+    ADD_OPTION_PTR(CU_JIT_INFO_LOG_BUFFER_SIZE_BYTES, (void *) buf_size);
+    ADD_OPTION_PTR(CU_JIT_ERROR_LOG_BUFFER, error_buf);
+    ADD_OPTION_PTR(CU_JIT_ERROR_LOG_BUFFER_SIZE_BYTES, (void *) buf_size);
+
+    PYTHON_FOREACH(key_value, py_options)
+      ADD_OPTION_PTR(
+          py::extract<CUjit_option>(key_value[0]), 
+          (void *) py::extract<int>(key_value[1])());
+#undef ADD_OPTION
+    
+    CUDAPP_PRINT_CALL_TRACE("cuModuleLoadDataEx");
+    CUresult cu_status_code; \
+    cu_status_code = cuModuleLoadDataEx(&mod, mod_buf, options.size(), 
+         const_cast<CUjit_option *>(&*options.begin()),
+         const_cast<void **>(&*option_values.begin()));
+
+    size_t info_buf_size = size_t(option_values[1]);
+    size_t error_buf_size = size_t(option_values[3]);
+
+    if (message_handler != py::object())
+      message_handler(cu_status_code == CUDA_SUCCESS,
+          std::string(info_buf, info_buf_size),
+          std::string(error_buf, error_buf_size));
+
+    if (cu_status_code != CUDA_SUCCESS)
+      throw cuda::error("cuModuleLoadDataEx", cu_status_code, 
+          std::string(error_buf, error_buf_size).c_str());
+#else
+    if (py::len(py_options))
+      throw cuda::error("module_from_buffer", CUDA_ERROR_INVALID_VALUE,
+          "non-empty options argument only supported on CUDA 2.1 and newer");
+
+    CUDAPP_CALL_GUARDED(cuModuleLoadData, (&mod, mod_buf));
+#endif
+
+    return new module(mod);
+  }
+
+  // }}}
+
+  PyObject *device_allocation_to_long(device_allocation const &da)
+  {
+    return PyLong_FromUnsignedLong((CUdeviceptr) da);
+  }
+
+  // {{{ pagelocked memory <-> numpy
+
+  py::handle<> pagelocked_empty(py::object shape, py::object dtype, 
+      py::object order_py, unsigned mem_flags)
+  {
+    PyArray_Descr *tp_descr;
+    if (PyArray_DescrConverter(dtype.ptr(), &tp_descr) != NPY_SUCCEED)
+      throw py::error_already_set();
+
+    py::extract<npy_intp> shape_as_int(shape);
+    std::vector<npy_intp> dims;
+
+    if (shape_as_int.check())
+      dims.push_back(shape_as_int());
+    else
+      std::copy(
+          py::stl_input_iterator<npy_intp>(shape),
+          py::stl_input_iterator<npy_intp>(),
+          back_inserter(dims));
+
+    std::auto_ptr<host_allocation> alloc(
+        new host_allocation(
+          tp_descr->elsize*pycuda::size_from_dims(dims.size(), &dims.front()),
+          mem_flags)
+        );
+
+    NPY_ORDER order = PyArray_CORDER;
+    PyArray_OrderConverter(order_py.ptr(), &order);
+
+    int ary_flags = 0;
+    if (order == PyArray_FORTRANORDER)
+      ary_flags |= NPY_FARRAY;
+    else if (order == PyArray_CORDER)
+      ary_flags |= NPY_CARRAY;
+    else
+      throw std::runtime_error("unrecognized order specifier");
+
+    py::handle<> result = py::handle<>(PyArray_NewFromDescr(
+        &PyArray_Type, tp_descr,
+        dims.size(), &dims.front(), /*strides*/ NULL,
+        alloc->data(), ary_flags, /*obj*/NULL));
+
+    py::handle<> alloc_py(handle_from_new_ptr(alloc.release()));
+    PyArray_BASE(result.get()) = alloc_py.get();
+    Py_INCREF(alloc_py.get());
+
+    return result;
+  }
+
+  // }}}
+
+  // }}}
+
+
+
+
   bool have_gl_ext()
   {
 #ifdef HAVE_GL
@@ -410,6 +433,8 @@ BOOST_PYTHON_MODULE(_driver)
   py::def("get_driver_version", cuda::get_driver_version);
 #endif
 
+  // {{{ exceptions
+
 #define DECLARE_EXC(NAME, BASE) \
   Cuda##NAME = py::handle<>(PyErr_NewException("pycuda._driver." #NAME, BASE, NULL)); \
   py::scope().attr(#NAME) = Cuda##NAME;
@@ -427,6 +452,9 @@ BOOST_PYTHON_MODULE(_driver)
     py::register_exception_translator<cuda::error>(translate_cuda_error);
   }
 
+  // }}}
+
+  // {{{ constants
 #if CUDA_VERSION >= 2000
   py::enum_<CUctx_flags>("ctx_flags")
     .value("SCHED_AUTO", CU_CTX_SCHED_AUTO)
@@ -445,6 +473,9 @@ BOOST_PYTHON_MODULE(_driver)
   py::enum_<CUevent_flags>("event_flags")
     .value("DEFAULT", CU_EVENT_DEFAULT)
     .value("BLOCKING_SYNC", CU_EVENT_BLOCKING_SYNC)
+#if CUDA_VERSION >= 3020
+    .value("DISABLE_TIMING", CU_EVENT_DISABLE_TIMING)
+#endif
     ;
 #endif
 
@@ -463,7 +494,13 @@ BOOST_PYTHON_MODULE(_driver)
 #if CUDA_VERSION >= 3000
   {
     py::class_<array3d_flags> cls("array3d_flags", py::no_init);
+    // deprecated
     cls.attr("ARRAY3D_2DARRAY") = CUDA_ARRAY3D_2DARRAY;
+
+    cls.attr("2DARRAY") = CUDA_ARRAY3D_2DARRAY;
+#if CUDA_VERSION >= 3010
+    cls.attr("SURFACE_LDST") = CUDA_ARRAY3D_SURFACE_LDST;
+#endif
   }
 #endif
 
@@ -471,6 +508,9 @@ BOOST_PYTHON_MODULE(_driver)
     .value("WRAP", CU_TR_ADDRESS_MODE_WRAP)
     .value("CLAMP", CU_TR_ADDRESS_MODE_CLAMP)
     .value("MIRROR", CU_TR_ADDRESS_MODE_MIRROR)
+#if CUDA_VERSION >= 3020
+    .value("BORDER", CU_TR_ADDRESS_MODE_BORDER)
+#endif
     ;
 
   py::enum_<CUfilter_mode>("filter_mode")
@@ -524,6 +564,11 @@ BOOST_PYTHON_MODULE(_driver)
     .value("CONCURRENT_KERNELS", CU_DEVICE_ATTRIBUTE_CONCURRENT_KERNELS)
     .value("ECC_ENABLED", CU_DEVICE_ATTRIBUTE_ECC_ENABLED)
 #endif
+#endif
+#if CUDA_VERSION >= 3020
+    .value("PCI_BUS_ID", CU_DEVICE_ATTRIBUTE_PCI_BUS_ID)
+    .value("PCI_DEVICE_ID", CU_DEVICE_ATTRIBUTE_PCI_DEVICE_ID)
+    .value("TCC_DRIVER", CU_DEVICE_ATTRIBUTE_TCC_DRIVER)
 #endif
     ;
 
@@ -586,6 +631,9 @@ BOOST_PYTHON_MODULE(_driver)
     .value("COMPUTE_13", CU_TARGET_COMPUTE_13)
 #if CUDA_VERSION >= 3000
     .value("COMPUTE_20", CU_TARGET_COMPUTE_20)
+#endif
+#if CUDA_VERSION >= 3020
+    .value("COMPUTE_21", CU_TARGET_COMPUTE_21)
 #endif
     ;
 
@@ -604,6 +652,17 @@ BOOST_PYTHON_MODULE(_driver)
   }
 #endif
 
+#if CUDA_VERSION >= 3010
+  py::enum_<CUlimit>("limit")
+    .value("STACK_SIZE", CU_LIMIT_STACK_SIZE)
+    .value("PRINTF_FIFO_SIZE", CU_LIMIT_PRINTF_FIFO_SIZE)
+#if CUDA_VERSION >= 3020
+    .value("MALLOC_HEAP_SIZE", CU_LIMIT_MALLOC_HEAP_SIZE)
+#endif
+    ;
+#endif
+
+
   // graphics enums -----------------------------------------------------------
 #if CUDA_VERSION >= 3000
   py::enum_<CUgraphicsRegisterFlags>("graphics_register_flags")
@@ -626,10 +685,12 @@ BOOST_PYTHON_MODULE(_driver)
     ;
 #endif
 
+  // }}}
 
   py::def("init", init,
       py::arg("flags")=0);
 
+  // {{{ device
   {
     typedef device cl;
     py::class_<cl>("Device", py::no_init)
@@ -647,7 +708,9 @@ BOOST_PYTHON_MODULE(_driver)
           (py::args("self"), py::args("flags")=0))
       ;
   }
+  // }}}
 
+  // {{{ context
   {
     typedef context cl;
     py::class_<cl, shared_ptr<cl>, boost::noncopyable >("Context", py::no_init)
@@ -670,9 +733,25 @@ BOOST_PYTHON_MODULE(_driver)
 
       .def("get_current", (boost::shared_ptr<cl> (*)()) &cl::current_context)
       .staticmethod("get_current")
+
+#if CUDA_VERSION >= 3010
+      .DEF_SIMPLE_METHOD(set_limit)
+      .staticmethod("set_limit")
+      .DEF_SIMPLE_METHOD(get_limit)
+      .staticmethod("get_limit")
+#endif
+#if CUDA_VERSION >= 3020
+      .DEF_SIMPLE_METHOD(get_cache_config)
+      .staticmethod("get_cache_config")
+      .DEF_SIMPLE_METHOD(set_cache_config)
+      .staticmethod("set_cache_config")
+      .DEF_SIMPLE_METHOD(get_api_version)
+#endif
       ;
   }
+  // }}}
 
+  // {{{ stream
   {
     typedef stream cl;
     py::class_<cl, boost::noncopyable>
@@ -681,7 +760,9 @@ BOOST_PYTHON_MODULE(_driver)
       .DEF_SIMPLE_METHOD(is_done)
       ;
   }
+  // }}}
 
+  // {{{ module
   {
     typedef module cl;
     py::class_<cl, boost::noncopyable, shared_ptr<cl> >("Module", py::no_init)
@@ -691,6 +772,11 @@ BOOST_PYTHON_MODULE(_driver)
       .def("get_texref", module_get_texref, 
           (py::args("self", "name")),
           py::return_value_policy<py::manage_new_object>())
+#if CUDA_VERSION >= 3010
+      .def("get_surfref", module_get_surfref, 
+          (py::args("self", "name")),
+          py::return_value_policy<py::manage_new_object>())
+#endif
       ;
   }
 
@@ -702,6 +788,9 @@ BOOST_PYTHON_MODULE(_driver)
        py::arg("message_handler")=py::object()),
       py::return_value_policy<py::manage_new_object>());
 
+  // }}}
+
+  // {{{ function
   {
     typedef function cl;
     py::class_<cl>("Function", py::no_init)
@@ -726,6 +815,10 @@ BOOST_PYTHON_MODULE(_driver)
       ;
   }
 
+  // }}}
+
+  // {{{ pointer holder
+
   {
     typedef device_allocation cl;
     py::class_<cl, boost::noncopyable>("DeviceAllocation", py::no_init)
@@ -737,12 +830,27 @@ BOOST_PYTHON_MODULE(_driver)
     py::implicitly_convertible<device_allocation, CUdeviceptr>();
   }
 
+  // }}}
+
+  {
+    typedef pointer_holder_base cl;
+    py::class_<pointer_holder_base_wrap, boost::noncopyable>(
+        "PointerHolderBase")
+      .def("get_pointer", py::pure_virtual(&cl::get_pointer))
+      ;
+
+    py::implicitly_convertible<pointer_holder_base, CUdeviceptr>();
+  }
+
   {
     typedef host_allocation cl;
     py::class_<cl, boost::noncopyable>("HostAllocation", py::no_init)
       .DEF_SIMPLE_METHOD(free)
 #if CUDA_VERSION >= 2020
       .DEF_SIMPLE_METHOD(get_device_pointer)
+#endif
+#if CUDA_VERSION >= 3020
+      .DEF_SIMPLE_METHOD(get_flags)
 #endif
       ;
   }
@@ -943,6 +1051,17 @@ BOOST_PYTHON_MODULE(_driver)
       ;
   }
 
+#if CUDA_VERSION >= 3010
+  {
+    typedef surface_reference cl;
+    py::class_<cl, boost::noncopyable>("SurfaceReference", py::no_init)
+      .DEF_SIMPLE_METHOD(set_array)
+      .def("get_array", &cl::get_array,
+          py::return_value_policy<py::manage_new_object>())
+      ;
+  }
+#endif
+
   py::scope().attr("TRSA_OVERRIDE_FORMAT") = CU_TRSA_OVERRIDE_FORMAT;
   py::scope().attr("TRSF_READ_AS_INTEGER") = CU_TRSF_READ_AS_INTEGER;
   py::scope().attr("TRSF_NORMALIZED_COORDINATES") = CU_TRSF_NORMALIZED_COORDINATES;
@@ -955,3 +1074,5 @@ BOOST_PYTHON_MODULE(_driver)
   pycuda_expose_gl();
 #endif
 }
+
+// vim: foldmethod=marker
