@@ -14,7 +14,6 @@ def have_pycuda():
     except:
         return False
 
-
 if have_pycuda():
     import pycuda.gpuarray as gpuarray
     import pycuda.driver as drv
@@ -23,6 +22,13 @@ if have_pycuda():
 
 
 
+def has_double_support():
+    from pycuda.driver import Context
+    return Context.get_device().compute_capability >= (1,3)
+
+
+
+
 class TestGPUArray:
     disabled = not have_pycuda()
 
@@ -211,13 +217,98 @@ class TestGPUArray:
     @mark_cuda_test
     def test_random(self):
         from pycuda.curandom import rand as curand
-        for dtype in [numpy.float32, numpy.float64]:
+
+        if has_double_support():
+            dtypes = [numpy.float32, numpy.float64]
+        else:
+            dtypes = [numpy.float32]
+
+        for dtype in dtypes:
             a = curand((10, 100), dtype=dtype).get()
 
             assert (0 <= a).all()
             assert (a < 1).all()
 
 
+    @mark_cuda_test
+    def test_array_gt(self):
+        """Test whether array contents are > the other array's
+        contents"""
+
+        a = numpy.array([5,10]).astype(numpy.float32)
+        a_gpu = gpuarray.to_gpu(a)
+        b = numpy.array([2,10]).astype(numpy.float32)
+        b_gpu = gpuarray.to_gpu(b)
+        result = (a_gpu > b_gpu).get()
+        assert result[0] == True
+        assert result[1] == False
+
+    @mark_cuda_test
+    def test_array_lt(self):
+        """Test whether array contents are < the other array's
+        contents"""
+
+        a = numpy.array([5,10]).astype(numpy.float32)
+        a_gpu = gpuarray.to_gpu(a)
+        b = numpy.array([2,10]).astype(numpy.float32)
+        b_gpu = gpuarray.to_gpu(b)
+        result = (b_gpu < a_gpu).get()
+        assert result[0] == True
+        assert result[1] == False
+
+    @mark_cuda_test
+    def test_array_le(self):
+        """Test whether array contents are <= the other array's
+        contents"""
+
+        a = numpy.array([5,10, 1]).astype(numpy.float32)
+        a_gpu = gpuarray.to_gpu(a)
+        b = numpy.array([2,10, 2]).astype(numpy.float32)
+        b_gpu = gpuarray.to_gpu(b)
+        result = (b_gpu <= a_gpu).get()
+        assert result[0] == True
+        assert result[1] == True
+        assert result[2] == False
+
+    @mark_cuda_test
+    def test_array_ge(self):
+        """Test whether array contents are >= the other array's
+        contents"""
+
+        a = numpy.array([5,10,1]).astype(numpy.float32)
+        a_gpu = gpuarray.to_gpu(a)
+        b = numpy.array([2,10,2]).astype(numpy.float32)
+        b_gpu = gpuarray.to_gpu(b)
+        result = (a_gpu >= b_gpu).get()
+        assert result[0] == True
+        assert result[1] == True
+        assert result[2] == False
+
+    @mark_cuda_test
+    def test_array_eq(self):
+        """Test whether array contents are == the other array's
+        contents"""
+
+        a = numpy.array([5,10]).astype(numpy.float32)
+        a_gpu = gpuarray.to_gpu(a)
+        b = numpy.array([2,10]).astype(numpy.float32)
+        b_gpu = gpuarray.to_gpu(b)
+        result = (a_gpu == b_gpu).get()
+        assert result[0] == False
+        assert result[1] == True
+
+    @mark_cuda_test
+    def test_array_ne(self):
+        """Test whether array contents are != the other array's
+        contents"""
+
+        a = numpy.array([5,10]).astype(numpy.float32)
+        a_gpu = gpuarray.to_gpu(a)
+        b = numpy.array([2,10]).astype(numpy.float32)
+        b_gpu = gpuarray.to_gpu(b)
+        result = (a_gpu != b_gpu).get()
+        assert result[0] == True
+        assert result[1] == False
 
 
     @mark_cuda_test
@@ -317,8 +408,13 @@ class TestGPUArray:
     def test_minmax(self):
         from pycuda.curandom import rand as curand
 
+        if has_double_support():
+            dtypes = [numpy.float64, numpy.float32, numpy.int32]
+        else:
+            dtypes = [numpy.float32, numpy.int32]
+
         for what in ["min", "max"]:
-            for dtype in [numpy.float64, numpy.float32, numpy.int32]:
+            for dtype in dtypes:
                 a_gpu = curand((200000,), dtype)
                 a = a_gpu.get()
 
@@ -335,7 +431,12 @@ class TestGPUArray:
         gran = 5
         l_m = l_a - l_a // gran + 1
 
-        for dtype in [numpy.float64, numpy.float32, numpy.int32]:
+        if has_double_support():
+            dtypes = [numpy.float64, numpy.float32, numpy.int32]
+        else:
+            dtypes = [numpy.float32, numpy.int32]
+
+        for dtype in dtypes:
             a_gpu = curand((l_a,), dtype)
             a = a_gpu.get()
 
@@ -430,6 +531,9 @@ class TestGPUArray:
     def test_astype(self):
         from pycuda.curandom import rand as curand
 
+        if not has_double_support():
+            return
+
         a_gpu = curand((2000,), dtype=numpy.float32)
 
         a = a_gpu.get().astype(numpy.float64)
@@ -450,8 +554,13 @@ class TestGPUArray:
     def test_complex_bits(self):
         from pycuda.curandom import rand as curand
 
+        if has_double_support():
+            dtypes = [numpy.complex64, numpy.complex128]
+        else:
+            dtypes = [numpy.complex64]
+
         n = 20
-        for tp in [numpy.complex64, numpy.complex128]:
+        for tp in dtypes:
             dtype = numpy.dtype(tp)
             from pytools import match_precision
             real_dtype = match_precision(numpy.dtype(numpy.float64), dtype)
@@ -464,8 +573,6 @@ class TestGPUArray:
             assert la.norm(z.get().conj() - z.conj().get()) == 0
 
 
-
-
 if __name__ == "__main__":
     # make sure that import failures get reported, instead of skipping the tests.
     import pycuda.autoinit
